

/*
   Main header file for FxEngine.
   FxEngine is a 3dengine using glide that works on a voodoo graphics
   compatible board, it works on voodoo graphics but voodoo2 is recommended
   if you want trilinear filtering and 800x600

   Currently supported features:
   - TRUE RGB lightning
   - TEXTURE MAPPING
   - TRILINEAR FILTERING
   - BILINEAR FILTERING
   - MIPMAPPING
   - W BUFFERING
   - UNLIMITED NUMBER OF LIGHTSOURCES
   - UNLIMITED NUMBER OF OBJECTS
   - UNLIMITED NUMBER OF POLYGONS
   - GOURAUD SHADING
*/

#define _FXENGINE_

#include <windows.h>
#include <glide.h>
#ifndef _FXDRIVER_
#include "fxedrv.h"
#endif

// defines:


// different texture types:

#define FXE_TEXTURE_MAP 0
#define FXE_TEXTURE_SPECIAL 1

// the different draw shapes:

#define FXE_SHAPE_POLYGONS 0
#define FXE_SHAPE_LINES 1
#define FXE_SHAPE_POINTS 2

// the different W buffering modes:

#define W_BUFFERING_ENABLED 0
#define W_BUFFERING_DISABLED 1

// the different resolutions:

#define FXE_RES_640x480 0
#define FXE_RES_800x600 1
#define FXE_RES_1024x768 2 
#define FXE_RES_1280x1024 3
#define FXE_RES_1600x1200 4

// the different rendering states:

#define FXE_RENDER_2D 0
#define FXE_RENDER_3D 1

// the supported object formats:

#define OBJECT_3DS 0

// valid sources from where to load objects

#define LOAD_FROM_FILE 0
#define LOAD_FROM_MEMORY 1

// The different buffers:

#define fxeBackBuffer 0
#define fxeFrontBuffer 1

// The different fogging modes:

#define NO_FOGGING 0
#define DISTANCE_FOGGING 1

// the different detaillevels:

#define LOW_DETAIL 0     // requires voodoo
						 // 640x480 and bilinear filtered textures

#define HIGH_DETAIL 1    // requires voodoo2
						 // 800x600 and trilinear filtered textures

#define MAXIMUM_DETAIL 2 // requires voodoo2 with SLI
                         // 1024x768 and trilinear filtered textures

#define AUTOMATIC_DETAIL 3      // selects detail level automaticly
					     									// depending on hardware present

// the different shading types:
#define GOURAUD_SHADING 1
#define PHONG_SHADING 2
#define FLAT_SHADING 3
#define GOURAUD_SHADING_WSH 4   // gouraud shading with specular highlights

// the different object types:
#define OBJECT_COMPLETELY_TEXTURED 0  // Every polygon is textured in the object
#define OBJECT_NOT_TEXTURED 1         // No polygon is textured in the object
#define OBJECT_COMBO 2                // Some polygons is textured and some are not

// Maximum number of textures in world:
// (Change this definition and rebuild FxEngine
// if you want more textures)
#define MAX_NUM_TEXTURES 500

// the different texturemodes
#define ONLY_TEXTURED 0
#define TEXTURE_MAPPED 1
#define NOT_TEXTURED 2
#define BUMP_MAPPED 4  
#define ENVIROMENT_MAPPED 8

// the different texture filtering modes
#define TRILINEAR_FILTERED 0
#define BILINEAR_FILTERED 1

#define FXETRUE 1
#define FXEFALSE 0


// globals


float HorRes, VerRes; // Horisontal and Vertical Resolution

float HalfHorRes, HalfVerRes;

// datatypes:

typedef int FxeDetailLevel;

typedef int FxeResolution;

typedef	int fxeVertexIndex;

typedef int fxePolygonIndex;


typedef struct fxeTextureCoordinate
{
  // Texture coordinates: upper left corner in texture
  // is 0,0 and lower right corner in texture is 1,1
  // Texture coordinates can be larger than 1.0. If so the
  // texture is tiled.

  float u,v;

} fxeTextureCoordinate;


typedef struct
{
	float x,y,u,v;
} fxeRectCoordinate;

typedef struct 
{
	fxeRectCoordinate RectCoordinates[4];
} fxeRectangle2d;


typedef struct fxeVector
{
  float x,y,z,w;
} fxeVector;

typedef struct fxeTexture
{

	// pointer to next texture in linked list.
  void *Next;
  
	// GrTexInfo structure of texture needed for glide 
	// calls
	GrTexInfo TexInfo;
  
	// Is this texture downloaded to texture ram
  char DownloadedToTRAM;

	// This flag is set to FXETRUE if a polygon is using 
	// this texture in the current frame
	char Used;

	// Number of polygons of the current frame that is 
	// using this texture
	int NumPolygons;

	// The number of bytes used by this texture
	int NumBytes;

	// Number of bytes reserved by this texture in 
	// Texture ram of 3dfx card.
	int NumBytesInTmu;   

  // The address of the texture in texture ram.
	FxU32 VideomemoryAddress;
  
	// a pointer to data of texture in system ram
	void *TextureData;

	// Type of texture, one of these:
	// FXE_TEXTURE_MAP
	// FXE_TEXTURE_SPECIAL
	char TexType;   

	// FXETRUE indices this texture is changed under a frame
	// and must be downloaded to a tmu. This is used for
	// animated textures and lightmapping code.
	// FXEFALSE indicates the texture haven't changed.
	char Changed;

	char *Filename;

} fxeTexture;

typedef fxeTexture *fxeTextureHandle;

typedef struct
{
	fxeTextureHandle LensTexture;
	// 1.0 here means the main lensflare, the lensflare
	// right from the lightsource. -1.0 means the oppocite
	// lensflare and values in between gives lensflares in 
	// between with 0.0 in the middle.
	float MulFactor;
	// The radius of the lensflare.
	float Radius; 
} fxeLensflare;

typedef struct fxeLensflares
{
	int Numflares;
	fxeLensflare *Lensfls;
} fxeLensflares;

typedef fxeLensflares *fxeLensflareHandle;


typedef struct
{
  void *Next;
  void *Prev;
  float x,y,z,r,g,b;
  float DistanceAffection;
	int NumLensflares;
	fxeLensflareHandle LFH;
	int Enabled;
	fxeVector Normal;

} fxeLightsource;





typedef struct fxeColor
{
	float r,g,b;
} fxeColor;

// A character type - with characters you can display
// counters and text, a must in video games.
// You can make your character translucent, some cool effects is
// possible. You can also change the size of the character, enabling
// some zooming effects. The best is that this is INCREDIBLE easy to
// implement with glide.

typedef struct fxeCharacter
{
  // Handle to texture of the character bitmap
  fxeTextureHandle TextureHandle;

  // Is the character translucent
  // FXETRUE if transparent and FXEFALSE if not
  int Translucent;

  // How much translucent is this character - a value from
  // 0 to 255.0
  float AlphaLevel;

  // Is the character Transparent, this means that
  // a color in the character shouldn't be drawn.
  // FXETRUE if transparent and FXEFALSE if not.
  int Transparent;

  // The color that is transparent:
  fxeColor TransparentColor;

  // Indicates that this character is used and contains a valid character.
  int Used;

} fxeCharacter, *fxeCharacterHandle;

typedef struct fxeCharacterset
{
  fxeCharacter Characterset[256];
  float Spacing;
  float CharWidth;
  float CharHeight;
  float AlphaLevel;
  int Translucent;
  int Transparent;
  fxeColor TransColor;
  int UseDefaults;

} fxeCharacterset, *fxeCharactersetHandle;


typedef struct fxeDrawVertex
{
  // x,y,z of vertex
  float x;    // 0
  float y;    // 4

  float z;    // 8

  float ooz;  // 12

  float q;    // 16

  // Vertex Color
  float r;    // 20
  float g;    // 24
  float b;    // 28

  // Texture Coordinates for tmu 0
  float u0,v0;  // 32

  // Texture Coordinates for tmu 1
  float u1,v1;  // 40

  // In how many polygons this vertex is present:
  int NumPolys;

  // pointer to list of polygon indices of polygons that have this
  // vertex in common.
  fxePolygonIndex *PolygonIndices;

	// Flag that indicates if this vertex must be performed
	// lightning calculations on.
	int Used;
} fxeDrawVertex;

typedef struct fxeVertex
{
  // x,y,z of vertex
  float x;  
  float y; 
  float z; 
} fxeVertex;

typedef struct
{

  fxeDrawVertex v0;
  fxeTextureCoordinate e0;
  fxeDrawVertex v1;
  fxeTextureCoordinate e1;

} fxeLine;



typedef struct
{

  fxeVector Normal;
  float Distance;

} fxePlane;

// all planes that make up view frustrum
fxePlane NearPlane;
fxePlane FarPlane;
fxePlane LeftPlane;
fxePlane RightPlane;
fxePlane UpPlane;
fxePlane DownPlane;
  

typedef struct fxePolygon
{
	// This color is always added to the polygon when rendering
	// independent of how much light the polygon is given.
	fxeColor AmbientColor; 
	
	// If this polygon is not texture mapped 
	// this is the diffusecolor that will be used to calculate the final
	// color of the polygon when taking into account the amount of light
	// on this polygon.

	fxeColor DiffuseColor;

  // if flat shading is used this is the flat shaded color -
  // only used internally by FxEngine
  fxeColor FlatShadedColor;

	// Pointer to list of indices into the polygons vertices
	fxeVertexIndex *VertexIndices;

	// Pointer to list of handles to clipped vertices,
	// These vertices are created by FxEngine when it clips
	// polygons to the viewing volume.
	//fxeDrawVertex *ClippedVertices;

  //fxeTextureCoordinate *ClippedEnvVer;

  int PartlyClipStatus;

	// AlphaLevel is set to the amount of transparency the polygon will have,
	// is only used if the Transparent variable is set to FXTRUE
	unsigned char AlphaLevel;

	// The handle to the texture that this polygon is using
	// (only used if TexMode is set to one of the texture alternatives.
	fxeTextureHandle TextureHandle;

  // A handle to a texture which is used for bump mapping:
  // this handle is only valid if the object this polygon resides in
  // have its BumpMapped flag set to true. The texture coordinates in
  // the bumpmap is the same as the texture coordinates.
  fxeTextureHandle BumpMapHandle;

  float BumpDepth;

	// Which texture mode the polygon is using
	// (NOT_TEXTURED,TEXTURE_MAPPED,BUMP_MAPPED,ENVIROMENT_MAPPED)
  // The texture modes can be or together to a texture mode
  // that is using texture mapping, bump mapping and enviroment
  // mapping
	int TexMode;

  // Pointer to list of texture coordinates
  fxeTextureCoordinate *TextureCoordinates;


	// Decides if the polygon is transparent or not - if this variable is
	// set to FXETRUE then the AlphaLevel variable is used to determine
	// the level of trancparency of this polygon.
	int Transparent;

  // Is the polygon clipped (3d volume clipped)
	// This variable is only used internally by FxEngine
	int Clipped;

  // Is the polygon backface removed
  // Only used internally by FxEngine
  int Culled;

  // Is a part of the polygon clipped - used when a polygon
  // is partly clipped to the viewing frustrum.
	int PartlyClipped;

  // Number of vertices including vertices added by the
  // polygon clipper
  int NumAddedVertices;

	// Is the polygon doublesided or can it be clipped
	// with backface removal
	int DoubleSided;

	// Number of vertices in polygon
	int NumVertices;

	// Normal vector to this polygon
	fxeVector Normal;

  float PolygonRadius;


} fxePolygon, *fxePolygonPtr;

typedef struct
{
  void *Next;
  void *Prev;
  fxePolygonPtr pPolygon;
  float MinZ;
} fxePolygonList;

typedef struct
{
  unsigned long NumPolys;
  unsigned long FrameCount;
} fxePerformance;

typedef struct fxeObject
{
	void *Prev;
	void *Next;
  fxePolygon *PolygonList;   	 // objects polygons
	fxeVertex *VertexList;       // local vertices

	fxeVertex *VerticesWorld;  // vertices transformed to world space
	fxeDrawVertex *VerticesCamera; // vertices transformed to camera space
  fxeTextureCoordinate *EnviromentVerts;  // enviroment mapping vertices
	fxeVector *VertexNormals;	 // Normals to all vertices in object normalized to 1 in lenght:
	fxeColor *DynColors;       // Dynamic lightning colors.
                            
				   													 // - vertices ready to draw to screen
  float x,y,z;		   							 	 // position in world
	int Clipped;		   								 // Is the object clipped by FxEngine - only used internally

  float ObjectRadius;                // radius of the object
	int NumPolys;	           					 // Number of polygons in object
  int NumVertices;	   						 	 // Number of vertices in object
  int ShadingType;                   // The shadingtype the object uses
                                     // I have put the shadingtype in the
                                     // object structure instead of the
                                     // polygon structure. It's more
                                     // efficient and you probably never
                                     // want to use different
                                     // shadingtypes in a object.


  int ObjectType;                    // is the object completely texturemapped
                                     // or completely not texture mapped
                                     // or is it texture mapped and not
                                     // texturemapped. This variable is
                                     // used to increase speed as we don't
                                     // need to setup glide parameters every
                                     // polygon when rendering.
  int CompletelyVisible;             // FXETRUE if all polygons in object are
                                     // visible and FXEFALSE if not.
  float SpecConstant;
  int Transparent;
  float AlphaLevel;
  int LowAccuracyLightning;
  int BumpMapped;                    // FXETRUE if object is bump mapped, FXEFALSE
                                     // if not.

  int EnviromentMapped;              // FXETRUE if objects is enviroment mapped,
                                     // FXEFALSE if not

  fxeTextureHandle ETH;              // TextureHandle to the enviroment map
                                     // texture

  float EnvTexFactor;                // Value from 0.0 .. 1.0 describing the
                                     // the amount of alpha for Enviroment map:
                                     // specifying 0.0 makes the enviroment map
                                     // invisible and 1.0 makes the texture map
                                     // invisible

  int DrawShape;

  fxePolygonList *pPolygonSortList;  // Linked list of translucent polygons
                                     // sorted by Z value.


} fxeObject, *fxeObjectPtr;

typedef  fxeObject *fxeObjectHandle;

typedef struct
{
  // camera position
  float x,y,z;

  // camera angle
  float ax,ay,az;

  // If camera is moving this is how many frames we have left
  // to move and rotate it.

  int NumFramesLeft;

  // These variables tells how much the camera will move/rotate each frame
  float IncX;
  float IncY;
  float IncZ;
  float RotX;
  float RotY;
  float RotZ;

  // If the camera is set to follow a object this is the handle
  // to that object.
	fxeVertex PointToFollow;
	int FollowPoint;
} fxeCamera, *fxeCameraHandle;

fxeCameraHandle ActiveCamera;

typedef float fxeMatrix[4][4];


typedef fxeLightsource *fxeLightsourceHandle;

// All lightsources in world
fxeLightsource *Lightsources;

// functions:

// Returns 0xFFFFFFFF if failure and address in tmu if success. 
FxU32 fxeTMUAllocateMemory(GrChipID_t tmu, FxU32 Bytes);

void fxeResolutionGet(float *HorisontalRes, float *VerticalRes);

void fxeRenderWindowSet(int x, int y, int Width, int Height);

float VectorLenght(fxeVector *Vector);


void CrossProduct3d(fxeVector *Result, fxeVector *Vector1,
  fxeVector *Vector2);

// This function can be used to convert angle in degrees to and angle
// in radians.
float DegToRad(float Degree);

// This function returns the version of FxEngine. MainVersion is
// a pointer to a integer that will retrieve the mainversion of
// FxEngine. UnderVersion is a pointer to a integer that will retrieve
// the underversion of FxEngine.
void fxeGetVersion(int *MainVersion, int *UnderVersion);

// call this function first - initializes FxEngine and sets 3dfx
// videomode.
// Returns FXETRUE if successful and FXEFALSE if failure
int fxeInitialize(HWND hwnd, int ResToUse);

// This function should be called before your application terminates
// frees memory used by FxEngine and restores videomode to videomode
// used before fxeInitialize was called.
void fxeShutdown(void);


// Camera related functions:

// Moves the camera to the position in world specified by the x,y,z
// arguments.
void fxeCameraMove(float x, float y, float z);

// Rotates the camera to the angle specified by the xa,ya,za arguments.
void fxeCameraRotate(float xa, float ya, float za);


// Ambient light related functions:

void fxeAmbientlightLight(float r, float g, float b);


// Lightsource related functions:

// Creates a lensflare which you can bind to one or more
// lightsources.
fxeLensflareHandle fxeLensflareCreate(int NumFlares, fxeTextureHandle *TextureHandles,
	float *Radiuses, float *MulFactors);

// To create a lightsource - call this function. It returns a handle
// to a lightsource which can be used in other lightsource related
// functions


fxeLightsourceHandle fxeLightsourceAdd(float r, float g, float b,
  float x, float y, float z, float DistanceAffection,
	fxeLensflareHandle LFH, fxeVector *pNormal);

// Moves the lightsource with handle LightsourceHandle
// to the position x,y,z in world.
void fxeLightsourceMove(fxeLightsourceHandle LightsourceHandle,
	float x, float y, float z);

// Sets the rgb components of the lightsource with handle
// LightsourceHandle. The rgb components of the lightsource is given
// in the r,g,b arguments.
void fxeLightsourceLight(fxeLightsourceHandle LightsourceHandle,
	float r, float g, float b);

// Deletes the lightsource with handle LightsourceHandle from the world.
void fxeLightsourceDelete(fxeLightsourceHandle LightsourceHandle);


// Texture related functions:

// Loads a 3df texture file from disk, returns
// NULL if failure and handle to texture if
// successful. Returns a handle to the loaded texture.
fxeTextureHandle fxeTextureLoad(char *Filename, char TexType);

fxeTextureHandle fxeTextureLoadToMemory(char *filename);


// Deletes the texture with handle TextureHandle.
void fxeTextureDelete(fxeTextureHandle TextureHandle);


// Object related functions:

void fxeObjectAddVertex(fxeObjectHandle ObjectHandle, float x, float y,
  float z);

void fxeObjectAddPolygon(fxeObjectHandle ObjectHandle, int VertexIndex1,
  int VertexIndex2, int VertexIndex3, int DoubleSidedState,
  fxeTextureHandle TH, float u0, float v0, float u1, float v1, float
  u2, float v2);

void fxeObjectPrecalculateNormals(fxeObjectHandle ObjectHandle);

void fxeObjectPerformPrecalculations(fxeObjectHandle ObjectHandle, float
  ObjectSize);

fxeObject * fxeObjectInitDefaultState(void);

// Loads a object from disk and positions it at 0,0,0
// the loader can only load 3ds files for the moment.
// The 3ds loader loads a entire scene with multiple objects
// and makes these objects to a big object.
fxeObject * fxeObjectLoad(char *Filename, int FileFormat,
  float ScaleFactor, int UseMaterials);

fxeObject * fxeObjectLoadFromExe(int ResID, int FileFormat,
  float ScaleFactor, int UseTextures);

void fxeObjectSetShape(fxeObjectHandle ObjectHandle, int Shape);

void fxeObjectSetEnviromentAlpha(fxeObjectHandle ObjectHandle,
  float EnviromentAlphaLevel);

// Adds the object whose pointer is sent in the Object argument
// to the worlds objects.
// The function returns a handle to the object.
// This handle can be used in other object related functions.
fxeObjectHandle fxeObjectAdd(fxeObject *Object);

// Rotates a the object with handle ObjectHandle
// relative to its own coordinate system. The rotation angles is
// specified in the xa,ya,za arguments.
void fxeObjectRotate(fxeObjectHandle ObjectHandle,
  float xa, float ya, float za);

void fxeObjectSetShadingType(fxeObjectHandle ObjectHandle, int ShadingType);

void fxeObjectMakeEnviromentMapped(fxeObjectHandle OH,
  fxeTextureHandle TH, float AlphaLevel);

// Moves the object with handle ObjectHandle to the new position in
// world specified by the x,y,z arguments.
void fxeObjectMove(fxeObjectHandle ObjectHandle,
  float x, float y, float z);

// Scales the object with handle ObjectHandle to a new size,
// the function assumes the object is centered around its own
// coordinate system.
void fxeObjectScale(fxeObjectHandle ObjectHandle, float SFactor);

// Deletes the object with handle ObjectHandle from the world.
void fxeObjectDelete(fxeObjectHandle ObjectHandle);


// Rendering associated functions:

// Renders the world. This function shall be called when the scene
// shall be rendered. The DisplayWorld parameter specifies if the
// world shall be displayed or not. FXETRUE specifies that the world
// will be displayed and FXEFALSE specifies that the world will not
// be displayed.
void fxeWorldRender(int DisplayWorld);

// This function displays the rendered World. This function needs to be called
// if DisplayWorld parameter in the call fxeWorldRender was FXEFALSE or else
// the world will not be displayed or the screen.
void fxeWorldDisplay(void);

// Character associated functions:

fxeCharactersetHandle fxeCharactersetLoad(char *FilePrefix,
  int UseDefaultSettings, int Translucent,
  float AlphaLevel, int Transparent, fxeColor *TransColor, float Spacing,
  float CharWidth, float CharHeight);


void fxeSetDefaultCharacterset(fxeCharactersetHandle CSH);

void fxeCharactersetSetDefaultDrawingMode(fxeCharactersetHandle CSH);

void fxeCharactersetSetSpecialDrawingMode(fxeCharactersetHandle CSH,
  int Translucent, float AlphaLevel, int Transparent, fxeColor *TransColor,
  float Spacing, float CharacterWidth, float CharacterHeight);

void fxeStringDrawXY(char *String, float X, float Y);

fxeCharacterHandle fxeBitmapCreate(fxeTextureHandle TextureHandle,
  int Translucent, float AlphaLevel, int Transparent,
  fxeColor* TransparentColor);

void fxeBitmapDrawXY(fxeCharacterHandle CharHandle, float X, float Y,
  float Width, float Height);

// not to be called by application - use fxeStringDrawXY instead
void fxeCharacterDraw(fxeCharacterHandle CharHandle, float X, float Y,
  float Width, float Height);


void fxeFogSetMode(int FogMode);

void fxeFogSetColor(fxeColor *FogColor);

void fxeStartFrameCount(void);

// this function returns the number of frames that have
// been rendered since last call to fxeStartFrameCount
unsigned long fxeGetFrameCount(void);

void fxeBackgroundSetColor(fxeColor *pColor);

void fxeBackgroundClear(void);

void fxeSetRenderState(int RenderState);

void fxePerformanceInit(void);

fxePerformance * fxePerformanceGet(void);

void fxeVSyncEnable(void);

void fxeVSyncDisable(void);

void fxeObjectMakeTranslucent(fxeObjectHandle ObjectHandle, float AlphaLevel);

void fxeObjectDisableTranslucency(fxeObjectHandle ObjectHandle);

void fxeDriverGetInfo(fxeDriverInfo *DriverInfo);

void fxeLooseFocus(void);

void fxeSetFocus(void);

void fxeFarClippingPlaneSet(float Distance);

void fxeNearClippingPlaneSet(float Distance);

void fxeCameraSetActive(fxeCameraHandle CH);

fxeCameraHandle fxeCameraCreate(float wx, float wy, float wz,
  float ax, float ay, float az);

void fxeCameraMoveIncremental(int NumFrames,
  fxeVector *pMoveVector, float RotX, float RotY, float RotZ);

void fxeCameraEnableSyncToObject(fxeObjectHandle OH);

void fxeCameraDisableSyncToObject(void);

int fxeObjectDetectCollision(fxeObjectHandle OH1, fxeObjectHandle OH2);

void MatXVer(fxeMatrix Matrix, fxeVertex *Vertex, fxeVertex *Result);
void MatXVerToDrawVer(fxeMatrix Matrix, fxeVertex *Vertex, fxeDrawVertex *Result);
void MatXDrawVerToDrawVer(fxeMatrix Matrix, fxeDrawVertex *Vertex, fxeDrawVertex *Result);

int fxeHaveBuiltLightmaps(void);

void ClipTextureRectangle2d(fxeRectangle2d *pRec2d);

void TransformToWorld(fxeObjectHandle ObjectHandle);
