
/*
 *  3D Studio object reader
 */

/*
    Original source modified by Andreas Ingo
    - converted to c
    - added loading of materials and textures
*/



#define FILELOAD
#define MEMLOAD

#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include "3dsload.h"


static void  (* dread)  (void * dest, int len);
static void  (* dsetpos)(dword pos);
static dword (* dgetpos)(void);

static jmp_buf EnvState;

int ColorFound;

H3dsMaterial *CurrMaterial;


#ifdef FILELOAD
#include <stdio.h>

static FILE * InFile=0;

static void FileRead(void * dest, int len)
{
	if(fread(dest, len, 1, InFile) != 1) {
		//printf("Error reading file\n");
		longjmp(EnvState, 1);
	}
}

static void FileSetpos(dword pos)
{
	if(fseek(InFile, (long)pos, SEEK_SET) != 0) {
		//printf("Error moving filepointer\n");
		longjmp(EnvState, 1);
	}
}

static dword FileGetpos(void)
{
	long pos;
	if((pos=ftell(InFile)) == -1L) {
		//printf("Error getting fileposition\n");
		longjmp(EnvState, 1);
	}
	return (dword) pos;
}
#endif

#ifdef MEMLOAD
static byte * MemBuffer=0;
static dword MemBufferLen=0, MemBufferIndex=0;

static void MemRead(void * dest, int len)
{
	// Never allow any reads past the membuffer.
	if(MemBufferIndex+len > MemBufferLen) {
		//printf("Bad memread\n");
		longjmp(EnvState, 1);
	}
	memcpy(dest, &MemBuffer[MemBufferIndex], len);
	MemBufferIndex+=len;
}

static void MemSetpos(dword pos)
{
	// It should be legal to move the position one byte
	// past the membuffer. The mem must not be read though.
	if(pos > MemBufferLen) {
		//printf("Bad memsetpos\n");
		longjmp(EnvState, 1);
	}
	MemBufferIndex=pos;
}

static dword MemGetpos(void)
{
	return MemBufferIndex;
}
#endif



static H3dsScene * Scene;

H3dsMeshObj * GetMeshObj(void)
{
	void * mem;
  H3dsMeshObj * obj;

	if((mem=realloc(Scene->meshobjlist,
					sizeof(H3dsMeshObj)*(Scene->meshobjs+1))) == 0) {
		//printf("Error reallocating mem\n");
		longjmp(EnvState, 1);
	}
	Scene->meshobjlist=(H3dsMeshObj *) mem;
  obj=&Scene->meshobjlist[Scene->meshobjs++];
	memset(obj, 0, sizeof(H3dsMeshObj));
  obj->CurrFaceMaterial = &obj->FaceMaterials;
	return obj;
}

static void * getmem(int size)
{
	void * mem;
	if((mem=malloc(size))==0) {
		//printf("Error allocating mem\n");
		longjmp(EnvState, 1);
	}
	return mem;
}

// Each 3DS data-chunk starts with a 6 byte header.
// The first item in the header is a 2 byte (word) id-number.
// After that follows a dword wich gives the size of
// the data-chunk including the header. The size can be used
// as an relative offset to the next chunk.

// tab 4
enum {
  CHUNK_RGB1      	= 0x0011,  // 3 bytes of RGB
  CHUNK_RGB2      	= 0x0012,  // 3 floats of RGB
//  CHUNK_RBGB2     	= 0x0012,	// ?? NOT HLS.
  CHUNK_AMOUNT          = 0x0030,
	CHUNK_MAIN      	= 0x4D4D,
     CHUNK_EDIT3DS    = 0x3D3D,
      CHUNK_BKGCOLOR    = 0x1200,
	  CHUNK_AMBCOLOR  	= 0x2100,
	  CHUNK_OBJBLOCK  	= 0x4000,
	   CHUNK_TRIMESH   	= 0x4100,
		CHUNK_VERTLIST  = 0x4110,
		CHUNK_FACELIST  = 0x4120,
		CHUNK_FACEMAT   = 0x4130,
		CHUNK_MAPLIST   = 0x4140,
		CHUNK_SMOOLIST  = 0x4150,
		CHUNK_TRMATRIX  = 0x4160,
	  CHUNK_LIGHT     	= 0x4600,
		CHUNK_SPOTLIGHT = 0x4610,
	  CHUNK_CAMERA    	= 0x4700,
		CHUNK_MATERIAL  	= 0xAFFF,
	  CHUNK_MATNAME   	= 0xA000,
	  CHUNK_AMBIENT   	= 0xA010,
	  CHUNK_DIFFUSE   	= 0xA020,
	  CHUNK_SPECULAR  	= 0xA030,
    CHUNK_TWOSIDED    = 0xA081,
		CHUNK_TRANSPARENCY    = 0xA050,
	  CHUNK_TEXTURE   	= 0xA200,
	  CHUNK_BUMPMAP   	= 0xA230,
	  CHUNK_MAPFILE   	= 0xA300,
	 CHUNK_KEYFRAMER 	= 0xB000,
	  CHUNK_FRAMES      = 0xB008
};


static void ReadVertList(H3dsMeshObj * meshobj)
{
	word nv;
  int k,n ;
	dread(&nv, sizeof(nv));
	meshobj->verts=nv;
  k=nv;
	meshobj->vertlist=(H3dsVert *) getmem(sizeof(H3dsVert)*k);
  for(n=0; n<k; n++) {
		dread(&meshobj->vertlist[n], sizeof(float32)*3);
	}
}

static void ReadFaceList(H3dsMeshObj * meshobj)
{
	word nv;
  int k,n;
	dread(&nv, sizeof(nv));
	meshobj->faces=nv;
  k=nv;
	meshobj->facelist=(H3dsFace *) getmem(sizeof(H3dsFace)*k);
  for(n=0; n<k; n++) {
		dread(&meshobj->facelist[n], sizeof(word)*4);

	}
}

static void ReadMapList(H3dsMeshObj * meshobj)
{
	word nv;
  int n;
  int k;

	dread(&nv, sizeof(nv));
	meshobj->maps=nv;
  k=nv;
	meshobj->maplist=(H3dsMap *) getmem(sizeof(H3dsMap)*k);
  for(n=0; n<k; n++) {
		dread(&meshobj->maplist[n], sizeof(float32)*2);
	}
}

static void ReadFaceMaterials(H3dsMeshObj * meshobj)
{
	word nv;
  int n;
  int k;
  char Matname[80];
  H3dsMaterial *pMaterial;
  int Exit;
  H3dsFaceMaterial *pPolygonMaterial;

  // Read in the name of the material:
  n=0;
	do {
    dread(&Matname[n++], 1);
  } while(Matname[n-1]!='\0' && n<sizeof(Matname));
  Matname[n-1]='\0';

  /*
     Search for a material with this Matname.
     (This assumes that all materials already have been loaded but
      I don't think this is a problem since materials is before the
      meshobjects in the file.)
  */

  pMaterial = &Scene->Materials;
  Exit = 0;

  while(pMaterial != NULL && Exit == 0)
  {
    if (pMaterial->Matname != NULL)
    {

      if (strcmp(pMaterial->Matname,Matname) == 0)
      {
        // Now we have found a material that matches the one found:

        meshobj->CurrFaceMaterial->pMaterial = pMaterial;

        Exit = 1;
      }
    }

    pMaterial = (H3dsMaterial*)pMaterial->Next;
  }

	dread(&nv, sizeof(nv));
  meshobj->CurrFaceMaterial->nPolygons=nv;
  k=nv;
  meshobj->CurrFaceMaterial->PolygonIndices=(int*) getmem(sizeof(int)*k);
  for(n=0; n<k; n++) {
    dread(&nv, sizeof(nv));
    meshobj->CurrFaceMaterial->PolygonIndices[n] = nv;
  }

  pPolygonMaterial = (H3dsFaceMaterial*)malloc(sizeof(H3dsFaceMaterial));
  pPolygonMaterial->Prior = meshobj->CurrFaceMaterial;
  pPolygonMaterial->Next = NULL;
  pPolygonMaterial->pMaterial = NULL;
  pPolygonMaterial->nPolygons = 0;
  pPolygonMaterial->PolygonIndices = NULL;
  meshobj->CurrFaceMaterial->Next = pPolygonMaterial;
  meshobj->CurrFaceMaterial = pPolygonMaterial;

}


static void ReadTraMatrix(H3dsMeshObj * meshobj)
{
	dread(&meshobj->TraMatrix, sizeof(float32)*12);
	meshobj->matrix=1;
}

static void ReadTriMeshBlocks(dword p, char * name)
{
	word id;
	dword len, pc;
	H3dsMeshObj * meshobj=GetMeshObj();
	strcpy(meshobj->name, name);
	while((pc=dgetpos()) < p)
	{
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((int)id)
		{
      case CHUNK_VERTLIST: ReadVertList (meshobj); break;
      case CHUNK_FACELIST: ReadFaceList (meshobj); break;
      case CHUNK_MAPLIST:  ReadMapList  (meshobj); break;
      case CHUNK_TRMATRIX: ReadTraMatrix(meshobj); break;
      case CHUNK_FACEMAT:  ReadFaceMaterials(meshobj); break;

			//case CHUNK_FACEMAT:
			//case CHUNK_SMOOLIST:
			default: dsetpos(pc+len);
		}
	}
}

static void ReadObjBlocks(dword p)
{
	word id;
	dword len, pc;
	char name[16];

	// The object name is the first item
	int n=0;
	do {
		dread(&name[n++], 1);
	} while(name[n-1]!='\0' && n<sizeof(name));
	name[n-1]='\0';

	while((pc=dgetpos()) < p)
	{
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((int)id)
		{
			case CHUNK_TRIMESH:	ReadTriMeshBlocks(pc+len, name); break;

			//case CHUNK_LIGHT:
			//case CHUNK_CAMERA:
			default: dsetpos(pc+len);
		}
	}
}

static void ReadMatName(void)
{
  // Reads the name of the material in into the current material in
  // the scene.
	int n=0;

  CurrMaterial->Matname = (char*)malloc(sizeof(char)*80);

	do {
    dread(&CurrMaterial->Matname[n++], 1);
  } while(CurrMaterial->Matname[n-1]!='\0' && n<80);
  CurrMaterial->Matname[n-1]='\0';
}

static void ReadTexFileName(void)
{
  // Reads the name of the material in into the current material in
  // the scene.
	int n=0;

  CurrMaterial->TexFilename = (char*)malloc(sizeof(char)*80);

	do {
    dread(&CurrMaterial->TexFilename[n++], 1);
  } while(CurrMaterial->TexFilename[n-1]!='\0' && n<80);
  CurrMaterial->TexFilename[n-1]='\0';
}

static void ReadTexture(dword p)
{
	word id;
	dword len, pc;

	while((pc=dgetpos()) < p)
	{
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((int)id)
		{

      case CHUNK_MAPFILE: ReadTexFileName(); break;
			default: dsetpos(pc+len);
		}
	}
}

void InitMaterial(H3dsMaterial *pMaterial)
{
  pMaterial->TexFilename = NULL;
  pMaterial->Matname = NULL;
  pMaterial->Prior = NULL;
  pMaterial->Next = NULL;
  pMaterial->TH = NULL;
  pMaterial->DoubleSided = 0;
	// Default ambient color level is 0
	pMaterial->ambient.red = 0.0f;
	pMaterial->ambient.green = 0.0f;
	pMaterial->ambient.blue = 0.0f;
	// Default diffuse color level is maximum
	pMaterial->diffuse.red = 1.0f;
	pMaterial->diffuse.green = 1.0f;
	pMaterial->diffuse.blue = 1.0f;
	// Default alphalevel is 1 (totally visible)
	pMaterial->alpha = 1.0f;
}

void SetTwoSidedMaterial(dword p)
{
  CurrMaterial->DoubleSided = 1;
  dsetpos(p);
}


/* Reads in an RGB byte triplet */
void ReadRGBByte(dword p, H3dsRGBColor *rgb) {
  unsigned char red, green, blue;

  dread(&red, sizeof(red));
  dread(&green, sizeof(green));
  dread(&blue, sizeof(blue));
  rgb->red = ((float)(red))/255.0f;
  rgb->green = ((float)(green))/255.0f;
  rgb->blue = ((float)(blue))/255.0f;
	if (rgb->red != 0.0f || rgb->green != 0.0f ||
		rgb->blue != 0.0f)
		ColorFound = 1;

}

/* Reads in an RGB FP triplet */
void ReadRGBFloat(dword p, H3dsRGBColor *rgb) {

  dread(&rgb->red, sizeof(float32));
  dread(&rgb->green, sizeof(float32));
  dread(&rgb->blue, sizeof(float32));
}



/* Reads in a color assignment */
void ReadShadingChunk(dword p, H3dsRGBColor *rgb) {
  word id;
  dword len, pc;

  while((pc=dgetpos()) < p) {
    dread(&id, sizeof(id));
    dread(&len, sizeof(len));
    switch((unsigned int)id) {
      case CHUNK_RGB1: ReadRGBByte(pc+len, rgb); break;
      case CHUNK_RGB2: ReadRGBByte(pc+len, rgb); break;
      default: dsetpos(pc+len);
    }
  }
}



void ReadAmountChunk(word *amount) {
  word temp;

  dread(&temp, sizeof(temp));
  *amount = temp;
}

/*
 * I interpret the transparency information as the alpha
 * value for the entire material
 */
void ReadAlphaChunk(dword p, float32 *alpha) {
  word id;
  dword len, pc;
  word amount;

  while((pc=dgetpos()) < p) {
    dread(&id, sizeof(id));
    dread(&len, sizeof(len));
    switch((unsigned int)id) {
      case CHUNK_AMOUNT:
	ReadAmountChunk(&amount);
	break;
      default: 
	dsetpos(pc+len);
    }
  }

  *alpha = (float32)amount / 100.0f;
  if (*alpha < .01)
    *alpha = 1.0;

}  


  
static void ReadMaterial(dword p)
{
	word id;
	dword len, pc;
  H3dsMaterial *pMaterial;

	while((pc=dgetpos()) < p)
	{
    
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((int)id)
		{
      case CHUNK_MATNAME: ReadMatName(); break;
      case CHUNK_TWOSIDED: SetTwoSidedMaterial(pc+len); break;
      case CHUNK_TEXTURE: ReadTexture(pc+len); break;
			
			case CHUNK_AMBIENT:
				ReadShadingChunk(pc+len, &CurrMaterial->ambient);
				break;

			case CHUNK_DIFFUSE:
				ReadShadingChunk(pc+len, &CurrMaterial->diffuse);

			case CHUNK_TRANSPARENCY: 
				ReadAlphaChunk(pc+len, &CurrMaterial->alpha); break;

			default: dsetpos(pc+len);
		}
  }


  // Create a new material:
  pMaterial = (H3dsMaterial*)malloc(sizeof(H3dsMaterial));
  InitMaterial(pMaterial);
  pMaterial->Prior = CurrMaterial;
  CurrMaterial->Next = pMaterial;
  CurrMaterial = pMaterial;
}

static void ReadObjMeshBlocks(dword p)
{
	word id;
	dword len, pc;



	while((pc=dgetpos()) < p)
	{
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((int)id)
		{
			case CHUNK_OBJBLOCK: ReadObjBlocks(pc+len); break;
      case CHUNK_MATERIAL: ReadMaterial(pc+len); break;

      //case CHUNK_AMBCOLOR:
			//case CHUNK_BKGCOLOR:
			default: dsetpos(pc+len);
		}
	}
}

static void ReadMainBlocks(dword p)
{
	word id;
	dword len, pc;

	while((pc=dgetpos()) < p)
	{
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((int)id)
		{

      case CHUNK_EDIT3DS: ReadObjMeshBlocks(pc+len); break;


			//case CHUNK_KEYFRAMER:
			default: dsetpos(pc+len);
		}
	}
}


void HFree3dsScene(H3dsScene * scene)
{
  int n;
  H3dsMeshObj *mobj;

	if(scene) {
		if(scene->meshobjlist) {
      for(n=scene->meshobjs-1; n>=0; n--) {
        mobj = &scene->meshobjlist[n];
				if(mobj->maplist)  free(mobj->maplist);
				if(mobj->facelist) free(mobj->facelist);
				if(mobj->vertlist) free(mobj->vertlist);
			}
			free(scene->meshobjlist);
		}
		free(scene);
	}
}

H3dsScene * HRead3dsScene(void * ptr, int what, dword size)
{
  int retval;
	int Loop;
	H3dsFaceMaterial *CurrFaceMat;


	ColorFound = 0;

	if(ptr==0)
		return 0;

	#ifdef FILELOAD
	if(what==0) {
		// Load from file
		InFile=(FILE *) ptr;
		dread=FileRead;
		dsetpos=FileSetpos;
		dgetpos=FileGetpos;
	}
	else
	#endif
	#ifdef MEMLOAD
	if(what==1) {
		// Read from mem
		MemBufferIndex=0;
		MemBufferLen=size;
		MemBuffer=(byte *) ptr;
		dread=MemRead;
		dsetpos=MemSetpos;
		dgetpos=MemGetpos;
	}
	else
	#endif
	{
		return 0;
	}

	if((Scene=(H3dsScene *) malloc(sizeof(H3dsScene)))==0)
		return 0;
	memset(Scene, 0, sizeof(H3dsScene));

  CurrMaterial = &Scene->Materials;
  InitMaterial(CurrMaterial);

  retval = setjmp(EnvState);

	if(retval==0) {
		// Return address set, start loading 3DS data.
		word id;
		dword len, pc;
		pc=dgetpos();
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		if((int)id!=CHUNK_MAIN) {
			HFree3dsScene(Scene);
			//printf("Not 3ds data\n");
			return 0;
		}
		ReadMainBlocks(pc+len);
		if (ColorFound == 0) 
		{
			// No colors was found in the object
			// set all colors to default.
			for (Loop=0;Loop<Scene->meshobjs;Loop++)
			{
				CurrFaceMat = &Scene->meshobjlist[Loop].
					FaceMaterials;
				while (CurrFaceMat != NULL)
				{
					if (CurrFaceMat->pMaterial != NULL)
					{
						CurrFaceMat->pMaterial->ambient.red = 0.0f;
						CurrFaceMat->pMaterial->ambient.green = 0.0f;
						CurrFaceMat->pMaterial->ambient.blue = 0.0f;
						CurrFaceMat->pMaterial->diffuse.red = 1.0f;
						CurrFaceMat->pMaterial->diffuse.green = 1.0f;
						CurrFaceMat->pMaterial->diffuse.blue = 1.0f;
					}
				
					CurrFaceMat = CurrFaceMat->Next;
				}
			}
		}
	}
	else {
        // There was an error, free the allocated mem and return NULL.
		HFree3dsScene(Scene);
		return 0;
	}

	return Scene;
}
